import logging
import sys

from nehushtan.helper.CommonHelper import CommonHelper

logger_class = CommonHelper.class_with_class_path('nehushtan.logger.NehushtanLogger', 'NehushtanLogger')
logger = logger_class('test-for-autoload', [logging.StreamHandler(stream=sys.stdout)])
logger.info('miao')



import os
import time
from multiprocessing import Pool

from nehushtan.queue.NehushtanQueueDelegate import NehushtanQueueDelegate
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask
from nehushtan.queue.situation.NoNextTaskSituation import NoNextTaskSituation


class NehushtanQueue:
    def __init__(self, delegate: NehushtanQueueDelegate):
        self.delegate: NehushtanQueueDelegate = delegate
        self.current_workers = {}

    def get_logger(self):
        return self.delegate.logger

    def get_pool_capacity(self) -> int:
        """
        How many processes could run in the same time.
        Decided by the delegate
        """
        return self.delegate.read_config_of_delegate((NehushtanQueueDelegate.CONFIG_KEY_POOL_CAPACITY,), 1)

    def loop(self):
        self.delegate.when_loop_starts()

        with Pool(processes=self.get_pool_capacity()) as pool:
            while True:
                # Check existed workers' status before generate new workers
                self._refresh_count_of_current_workers()

                self.delegate.read_latest_command()

                # Check Command
                if self.delegate.should_terminate():
                    if self.delegate.should_wait_for_all_workers_before_terminating():
                        # wait for all worker processes done
                        self._refresh_count_of_current_workers(wait_till_less_than=1)
                        pass
                    break

                # Compare Pool Size and Capacity
                current_pool_size = len(self.current_workers.items())
                if current_pool_size >= self.get_pool_capacity():
                    self.delegate.when_pool_is_full()
                    if self.delegate.should_wait_when_pool_is_full():
                        # wait for any worker process done
                        self._refresh_count_of_current_workers(wait_till_less_than=self.get_pool_capacity())
                        pass
                    continue

                if not self.delegate.is_runnable():
                    self.delegate.when_loop_should_not_run()
                    continue

                try:
                    task = self.delegate.check_next_task()

                    if task.is_exclusive():
                        # now daemon should wait for the other tasks to be over
                        self._refresh_count_of_current_workers(wait_till_less_than=1)
                        pass

                    if not task.before_execute():
                        self.delegate.when_task_not_executable(task)
                        continue

                    task_async_process = pool.apply_async(
                        NehushtanQueue.embedded_task_execute,
                        args=(task, self.delegate)
                    )

                    self.current_workers[task.get_task_reference()] = task_async_process
                    self.delegate.when_worker_process_created(task.get_task_reference(), task_async_process.__str__())

                except NoNextTaskSituation:
                    self.delegate.when_no_task_to_do()
                    continue
                except Exception as loop_once_error:
                    self.delegate.when_loop_reports_error(loop_once_error.__str__())

        self.delegate.when_loop_terminates()

    def _refresh_count_of_current_workers(self, wait_till_less_than=0) -> int:
        if len(self.current_workers.items()) > 0:
            delete_keys = []
            for mapped_task_id, mapped_task_async_result in self.current_workers.items():
                try:
                    self.get_logger().debug('Checking Task Async Status', {'task_id': mapped_task_id})
                    if mapped_task_async_result.ready():
                        result = mapped_task_async_result.get(1)
                        if mapped_task_async_result.successful():
                            self.get_logger().info(
                                'Task Finished Successfully',
                                {"task_id": mapped_task_id, 'return': result}
                            )
                        else:
                            self.get_logger().warning(
                                'Task Finished Unsuccessfully',
                                {"task_id": mapped_task_id, 'return': result}
                            )
                        delete_keys.append(mapped_task_id)
                        self.delegate.when_worker_process_confirmed_dead(mapped_task_id, result)
                        # self.beat('whenChildProcessConfirmedDead',f'For Task {mapped_task_id}')
                except Exception as checking_async_exception:
                    self.get_logger().exception('Exception in checking async tasks', checking_async_exception)

            for delete_key in delete_keys:
                del self.current_workers[delete_key]
                self.get_logger().debug('remove removable finished tasks', {'task_id': delete_key})

        x = len(self.current_workers.items())
        if wait_till_less_than <= 0:
            # no need to wait, return
            return x
        else:
            if x < wait_till_less_than:
                # pool space is enough now
                return x
            else:
                # wait for processes to die
                self.get_logger().debug(f'Waiting for pool space. Now {x} out of {self.get_pool_capacity()} running.')
                time.sleep(2)
                return self._refresh_count_of_current_workers(wait_till_less_than=wait_till_less_than)

    @staticmethod
    def embedded_task_execute(embedded_task: NehushtanQueueTask, delegate: NehushtanQueueDelegate):
        """
        This works in WORKER
        """
        delegate.when_to_execute_task(embedded_task, os.getpid())
        try:
            embedded_task.execute()
        except Exception as e:
            delegate.when_task_raised_exception(embedded_task, e)
        delegate.when_task_executed(embedded_task, os.getpid())



import time
from abc import abstractmethod

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask


class NehushtanQueueDelegate:
    QUEUE_RUNTIME_COMMAND_PAUSE = "PAUSE"
    QUEUE_RUNTIME_COMMAND_CONTINUE = "CONTINUE"
    QUEUE_RUNTIME_COMMAND_STOP = "STOP"
    # QUEUE_RUNTIME_COMMAND_FORCE_STOP = "FORCE-STOP"
    # QUEUE_RUNTIME_COMMAND_RESTART = "RESTART"
    # QUEUE_RUNTIME_COMMAND_FORCE_RESTART = "FORCE-RESTART"

    CONFIG_KEY_POOL_CAPACITY = 'POOL_CAPACITY'

    def __init__(self, config_dictionary: dict = None, logger: NehushtanLogger = None):
        if config_dictionary is None:
            config_dictionary = {}
        self.config_dictionary = config_dictionary

        if logger is None:
            logger = NehushtanLogger(
                f"{self.__class__.__name__}-DefaultLogger",
                [NehushtanLogger.make_stdout_handler()],
                with_process_info=True,
                with_thread_info=True
            )
        self.logger = logger

        self.latest_command = NehushtanQueueDelegate.QUEUE_RUNTIME_COMMAND_CONTINUE

    def read_config_of_delegate(self, keychain: tuple, default):
        return CommonHelper.read_dictionary(self.config_dictionary, keychain, default)

    def get_configured_pool_capacity(self) -> int:
        return self.read_config_of_delegate((NehushtanQueueDelegate.CONFIG_KEY_POOL_CAPACITY,), 1)

    def read_latest_command(self) -> str:
        """
        Override this to fetch command from real pipe.
        This works in MASTER
        """
        self.latest_command = self.QUEUE_RUNTIME_COMMAND_CONTINUE
        return self.latest_command

    def when_loop_starts(self):
        """
        To be executed before loop cycle starts
        This works in MASTER
        """
        pass

    @abstractmethod
    def when_loop_reports_error(self, error_message: str):
        """
        This works in MASTER
        """
        pass

    def is_runnable(self):
        """
        If not runnable, the daemon loop would sleep.
        This works in MASTER
        """
        return self.latest_command == self.QUEUE_RUNTIME_COMMAND_CONTINUE

    def should_terminate(self):
        """
        Tell daemon loop to exit.
        This works in MASTER
        """
        return (
            self.QUEUE_RUNTIME_COMMAND_STOP,
            # self.QUEUE_RUNTIME_COMMAND_FORCE_STOP,
            # self.QUEUE_RUNTIME_COMMAND_RESTART,
            # self.QUEUE_RUNTIME_COMMAND_FORCE_RESTART,
        ).__contains__(self.latest_command)

    @abstractmethod
    def when_loop_terminates(self):
        """
        When the loop gets ready to terminate by shouldTerminate instructed, execute this
        This works in MASTER
        """
        pass

    @abstractmethod
    def when_loop_should_not_run(self):
        """
        Sleep for a certain while.
        This works in MASTER
        """
        pass

    @abstractmethod
    def check_next_task(self) -> NehushtanQueueTask:
        """
        This works in MASTER
        """
        pass

    @abstractmethod
    def when_no_task_to_do(self):
        """
        When the loop cannot check for a task to do next, execute this
        This works in MASTER
        """
        pass

    @abstractmethod
    def when_task_not_executable(self, task: NehushtanQueueTask):
        """
        this is done before fork in pooled style
        """
        pass

    @abstractmethod
    def when_to_execute_task(self, task: NehushtanQueueTask, pid: int):
        """
        SINCE 0.1.24 add parameter pid
        Note: Any exceptions should be caught inside.
        This works in WORKER
        """
        pass

    @abstractmethod
    def when_task_executed(self, task: NehushtanQueueTask, pid: int):
        """
        SINCE 0.1.24 add parameter pid
        Note: Any exceptions should be caught inside.
        This works in WORKER
        """
        try:
            if task.after_execute():
                self.logger.info('TASK FINISHED', {'task_id': task.get_task_reference(), 'pid': pid})
            else:
                self.logger.warning('TASK FINISHING UNWELL', {'task_id': task.get_task_reference(), 'pid': pid})
        except Exception as e:
            self.logger.exception(f'when_task_executed for task {task.get_task_reference()} on PID {pid}', e)

    @abstractmethod
    def when_task_raised_exception(self, task: NehushtanQueueTask, exception: Exception):
        """
        This works in WORKER
        """
        pass

    def when_pool_is_full(self):
        """
        This works in MASTER
        """
        self.logger.warning('Pool is full, sleep for a while', {'size': self.get_configured_pool_capacity()})
        # self.beat('whenPoolIsFull', 'Pool Full Wait')
        time.sleep(30)

    def when_worker_process_created(self, task_reference, note: str = ''):
        """
        This works in MASTER
        """
        pass

    def when_worker_process_confirmed_dead(self, task_reference, detail=None):
        """
        This works in MASTER
        """
        pass

    def should_wait_when_pool_is_full(self):
        """
        This works in MASTER
        """
        return True

    def should_wait_for_all_workers_before_terminating(self):
        """
        This works in MASTER
        """
        return True
